#ifndef ASTL_CURSOR_COPY
#define ASTL_CURSOR_COPY

#include <hash_map>
#include "safe_state.h" // template<> class safe_state;

// TODO: copie des tags
// Non test

// Algorithm:   ccopy (cursor copy)
// Description: copy depth-first range [first, last) into a DFA during ascending
//              stage of the depth-first traversal triming uneeded paths
// Input:       a DFA and two depth-first cursors
// Output:      initial copied state
// Remark:      processing of the initial state is an external operation left up
//              to the user. The algorithm simply returns the upper copied
//              transition source state.

template <class DFirstCursor, class DFA>
typename DFA::State ccopy(DFA &out, DFirstCursor first, DFirstCursor last = DFirstCursor())
{
  typedef hash_map<typename DFirstCursor::State, safe_state<typename DFA::State>, 
    identity<typename DFirstCursor::State> > mapper;
  mapper mapping;
  typename DFirstCursor::State i = first.src();

  while (first != last) {     // while not end of range
    while(first.forward());   // get down as much as possible

    // copy aim state if final:
    safe_state<typename DFA::State> p;
    if (first.aim_final()) {
      safe_state<typename DFA::State> &dest = mapping[first.aim()];
      if (dest == out.null_state) {
	dest = out.new_state();
	out.final(dest) = true;
      }
    }

    do {                      // while moving backward (pop) copy current
                              // transition (q, first.letter(), p) 
      // copy source state if needed:
      mapper::const_iterator dest = mapping.find(first.aim());
      if (dest != mapping.end()) p = (*dest).second;
      
      if (p != out.null_state || first.src_final()) {
	safe_state<typename DFA::State> &q = mapping[first.src()];
	if (q == out.null_state) {
	  q = out.new_state();
	  out.final(q) = first.src_final();
	}
	// copy transition:
	if (p != out.null_state) 
	  out.set_trans(q, first.letter(), p);
	
	// shift (q, first.letter(), p) => (q', first.letter()', q)
	p = q;
      }
    } while (! first.forward());
  }
  return mapping[i];
}

// Algorithm:   clone
// Description: copy depth-first range [first, last) into a DFA in descending
//              stage of the depth-first traversal making thus an exact copy of
//              the "composante connexe"
// Input:       two depth-first cursor and a DFA
// Output:      initial copied state
// Remark:      processing of the initial state is an external operation left up
//              to the user. The algorithm simply returns the first copied
//              transition source state.

  
template <class DFirstCursor, class DFA>
typename DFA::State clone(DFA &out, DFirstCursor first, DFirstCursor last = DFirstCursor())
{
  if (first == last)
    return out.null_state;
  hash_map<typename DFirstCursor::State, safe_state<typename DFA::State>,
    identity<typename DFirstCursor::State> > mapping;
  typename DFirstCursor::State i = first.src();

  while (first != last) {     // while not end of range

    // copy source state if needed:
    safe_state<typename DFA::State> &source = mapping[first.src()];
    if (source == out.null_state) {
      source = out.new_state();
      out.final(source) = first.src_final();
    }
    typename DFA::State q = source;

    do {                      // while moving forward
                              // copy current transition (q, first.letter(), p)
      // copy aim state if needed:
      safe_state<typename DFA::State> &p = mapping[first.aim()];
      if (p == out.null_state) {
	p = out.new_state();
	out.final(p) = first.aim_final();
      }

      // copy transition:
      out.set_trans(q, first.letter(), p);
      // shift (q, first.letter(), p) => (p, first.letter()', p')
      q = p;
    } while (first.forward());

    while (!first.forward());   // pop
  }
  return mapping[i];
}

#endif      










