#ifndef ASTL_CLASS_COMBINATORY_CURSORS
#define ASTL_CLASS_COMBINATORY_CURSORS

#include <vector>

// A TESTER

// Defines:
// 1. combination_cursor (forward cursor)
// 2. permutation_cursor (forward cursor)

// A combination_cursor simulates a DFA whose language is all combinations
// of the word 1234...n  (C(n,p))
//
// Constructor:
// 1. n, p

class combination_cursor
{
protected:
  vector<bool> visited;
  unsigned int current;
  int          remaining;

public:
  typedef combination_cursor self;
  typedef void Tag;
  typedef vector<bool> State;

  combination_cursor(unsigned int n, unsigned int p) 
    : visited(n, false), current(0), remaining(p)
    { }

  State src() const {
    return visited;
  }

  State aim() const {
    vector<bool> tmp = visited;
    tmp[current] = true;
    return tmp;
  }

  void forward() {
    --remaining;
    visited[current] = true;
  }

  bool forward(unsigned int letter) {
    if (visited[letter - 1] == false) {
      visited[letter - 1] = true;
      --remaining;
      return true;
    }
    return false;
  }

  bool first_transition() {
    for(vector<bool>::const_iterator i = visited.begin(); i != visited.end(); ++i)
      if (*i == false) {
	current = i - visited.begin();
	return true;
      }
    return false;
  }

  bool next_transition() {
    for(vector<bool>::const_iterator i = visited.begin() + current + 1; i != visited.end(); ++i)
      if (*i == false) {
	current = i - visited.begin();
	return true;
      }
    return false;
  }

  unsigned int letter() const {
    return current + 1;
  }

  unsigned int operator*() const {
    return letter();
  }

  bool src_final() const {
    return remaining < 1;
  }

  bool aim_final() const {
    return remaining < 2;
  }

  bool operator== (const self &x) const {
    return remaining == x.remaining && current == x.current && 
      visited[current] == x.visited[current];
  }

  bool find(unsigned int letter) {
    if (visited[letter - 1] == false) {
      current = letter - 1;
      return true;
    }
    return false;
  }
};

// A permutation_cursor simulates a DFA whose language is all permutations
// of the word 1234...n 
//
// Constructor:
// 1. n

class permutation_cursor
{
protected:
  vector<bool> visited;
  unsigned int current;
  int          remaining;

public:
  typedef permutation_cursor self;
  typedef void Tag;
  typedef vector<bool> State;

  permutation_cursor(unsigned int n)
    : visited(n, false), current(0), remaining(n)
    { }

  State src() const {
    return visited;
  }

  State aim() const {
    vector<bool> tmp = visited;
    tmp[current] = true;
    return tmp;
  }

  void forward() {
    --remaining;
    visited[current] = true;
  }

  bool forward(unsigned int letter) {
    if (visited[letter - 1] == false) {
      visited[letter - 1] = true;
      --remaining;
      return true;
    }
    return false;
  }

  bool first_transition() {
    for(vector<bool>::const_iterator i = visited.begin(); i != visited.end(); ++i)
      if (*i == false) {
	current = i - visited.begin();
	return true;
      }
    return false;
  }

  bool next_transition() {
    for(vector<bool>::const_iterator i = visited.begin() + current + 1; i != visited.end(); ++i)
      if (*i == false) {
	current = i - visited.begin();
	return true;
      }
    return false;
  }

  unsigned int letter() const {
    return current + 1;
  }

  bool src_final() const {
    return remaining == 0;
  }

  bool aim_final() const {
    return remaining == 1;
  }

  bool operator== (const self &x) const {
    return remaining == x.remaining && current == x.current && 
      visited[current] == x.visited[current];
  }

  bool find(unsigned int letter) {
    if (visited[letter - 1] == false) {
      current = letter - 1;
      return true;
    }
    return false;
  }
};

#endif // ASTL_CLASS_COMBINATORY_CURSORS
	
    
	 
    
	 
