#ifndef ASTL_CLASS_DEFAULT_CURSORS
#define ASTL_CLASS_DEFAULT_CURSORS

// Defines:
// 1. template default_transition_cursor (forward cursor adapter)
// 2. template default_state_cursor      (forward cursor adapter)

// NON TESTE !!!!!!!!!!!!

// A default_transition_cursor uses the transition labelled with default_letter
// when required transition is undefined.
//
// Instanciation parameters:
// 1. ForwardCursor: the cursor type to be adapted. 
//
// Requirements:
// None
//

template <class ForwardCursor>
class default_transition_cursor : public ForwardCursor
{
protected:
  unsigned int def_letter;

public:
  typedef ForwardCursor             super;
  typedef default_transition_cursor self;
  
  default_transition_cursor(const ForwardCursor &x, unsigned int default_letter)
    : super(x), def_letter(default_letter)
    { }

  bool forward(unsigned int letter) {
    return super::forward(letter) || super::forward(default_letter);
  }
    
  // The default transitions does not appear in the outgoing transition traversal:
  bool first_transition() {
    if (c.first_transition()) {
      while (c.letter() == default_letter)
	if (!c.next_transition()) return false;
      return true;
    }
    return false;
  }
    
  bool next_transition() {
    do
      if (!c.next_transition()) return false;
    while (c.letter() == default_letter);
    return true;
  }

  bool find(unsigned int letter) {
    return find(letter) || find(def_letter);
  }
};

// A default_state_cursor moves to default state when required transition is undefined
// (makes the DFA complete).
//
// Instanciation parameters:
// 1. ForwardCursor: the cursor type to be adapted. 
//
// Requirements:
// None
//
// Warning: this could make the automaton cyclic
//
// TODO:
// What should be done for find and outgoing transitions traversal

template <class ForwardCursor>
class default_state_cursor : public ForwardCursor
{
public:
  typedef typename ForwardCursor::State State;
  typedef ForwardCursor        super;
  typedef default_state_cursor self;
  
protected:
  State def_state;

public:
  default_state_cursor(const ForwardCursor &x, const State &default_state)
    : super(x), def_state(default_state)
    { }

  bool forward(unsigned int letter) {
    if (! super::forward(letter))
      *this = def_state;
    return true;
  }
};

// A substitute_state_cursor moves to default state when required transition is undefined
// a try to move forward from this state.
//
// Instanciation parameters:
// 1. ForwardCursor: the cursor type to be adapted. 
//
// Requirements:
// None
//
// Warning: this could make the automaton cyclic
//
// TODO:
// What should be done for find and outgoing transitions traversal

template <class ForwardCursor>
class substitute_state_cursor : public ForwardCursor
{
public:
  typedef typename ForwardCursor::State State;
  typedef ForwardCursor        super;
  typedef substitute_state_cursor self;
  
protected:
  State def_state;

public:
  substitute_state_cursor(const ForwardCursor &x = super(), const State &default_state = State())
    : super(x), def_state(default_state)
    { }

  bool forward(unsigned int letter) {
    if (! super::forward(letter)) {
      *this = def_state;
      return super::forward(letter);
    }
    return true;
  }

  self& operator= (const State &x) {
    super::operator=(x);
    return *this;
  }
};



#endif // ASTL_20_CLASS_DEFAULT_CURSORS

