#ifndef ASTL_CURSORS_HELPER_FUNCTIONS
#define ASTL_CURSORS_HELPER_FUNCTIONS

#include "cursor.h"
#include "set_operation.h"
#include "filter.h"
#include "combinatory.h"
#include "string_cursor.h"
#include "sigma_star.h"
#include "debug.h"
#include <astl.h>

template <class DFA>
forward_cursor<DFA> forwardc(const DFA &a, typename DFA::State q)
{
  return forward_cursor<DFA>(a, q);
}

template <class DFA>
forward_cursor<DFA> forwardc(const DFA &a) {
  return forward_cursor<DFA>(a);
}

template <class DFA>
forward_cursor_debug<DFA> debugc(const DFA &a) {
  return forward_cursor_debug<DFA>(a);
}

template <class ForwardCursor>
stack_cursor<ForwardCursor> stackc(const ForwardCursor &x) {
  stack_cursor<ForwardCursor> r(x);
  if (r.first_transition())
    return r;
  return stack_cursor<ForwardCursor>();
}

template <class ForwardCursor>
dfirst_cursor<stack_cursor<ForwardCursor> > dfirstc(const ForwardCursor &x) {
  return dfirst_cursor<stack_cursor<ForwardCursor> > (stackc(x));
}

template <class ForwardCursor1, class ForwardCursor2>
intersection_cursor<ForwardCursor1, ForwardCursor2>
intersectionc(const ForwardCursor1 &x1, const ForwardCursor2 &x2) {
  return intersection_cursor<ForwardCursor1, ForwardCursor2>(x1, x2);
}

template <class ForwardCursor1, class ForwardCursor2>
union_cursor<ForwardCursor1, ForwardCursor2>
unionc(const ForwardCursor1 &x1, const ForwardCursor2 &x2) {
  return union_cursor<ForwardCursor1, ForwardCursor2>(x1, x2);
}

template <class ForwardCursor, class ObjectFunction>
filter_out_cursor<ForwardCursor, ObjectFunction>
filter_outc(const ForwardCursor &x, const ObjectFunction &f = ObjectFunction()) {
  return filter_out_cursor<ForwardCursor, ObjectFunction>(x, f);
}

template <class ForwardCursor, class ObjectFunction>
filter_in_cursor<ForwardCursor, ObjectFunction>
filter_inc(const ForwardCursor &x, const ObjectFunction &f) {
  return filter_in_cursor<ForwardCursor, ObjectFunction>(x, f);
}

template <class ForwardCursor1, class ForwardCursor2>
diff_cursor<ForwardCursor1, ForwardCursor2>
diffc(const ForwardCursor1 &x1, const ForwardCursor2 &x2) {
  return diff_cursor<ForwardCursor1, ForwardCursor2>(x1, x2);
}

template <class ForwardCursor1, class ForwardCursor2>
sym_diff_cursor<ForwardCursor1, ForwardCursor2>
sym_diffc(const ForwardCursor1 &x1, const ForwardCursor2 &x2) {
  return sym_diff_cursor<ForwardCursor1, ForwardCursor2>(x1, x2);
}

template <class ForwardCursor1, class ForwardCursor2>
concatenation_cursor<ForwardCursor1, ForwardCursor2>
concatenationc(const ForwardCursor1 &x1, const ForwardCursor2 &x2) {
  return concatenation_cursor<ForwardCursor1, ForwardCursor2>(x1, x2);
}

template <class ForwardCursor>
not_cursor<ForwardCursor> notc(const ForwardCursor &x) {
  return not_cursor<ForwardCursor>(x);
}

permutation_cursor permutationc(unsigned int n) {
  return permutation_cursor(n);
}

combination_cursor combinationc(unsigned int n, unsigned int p) {
  return combination_cursor(n, p);
}

template <class InputIterator>
string_cursor<InputIterator> stringc(InputIterator first, InputIterator last) {
  return string_cursor<InputIterator>(first, last);
}

string_cursor<char*> stringc(const char *s) {
  return string_cursor<char*>(s);
}

string_cursor<unsigned char*> stringc(const unsigned char *s) {
  return string_cursor<unsigned char*>(s);
}

sigma_star_cursor< > sigma_starc() {
  return sigma_star_cursor< >();
}


#endif // ASTL_CURSORS_HELPER_FUNCTIONS
