#ifndef ASTL_CLASS_LAZY_CURSOR
#define ASTL_CLASS_LAZY_CURSOR

#include <map>
#include "safe_state.h"
#include "cursor.h"

// NON TESTE
// NON OPTIMAL (first et next_transition)

// Defines:
// 1. lazy_cursor (forward cursor adapter)

// A lazy cursor builds an automaton incrementally

// Instanciation parameters:
// 1. ForwardCursor: the type of the cursor to be adapted
// 2. DFA: a ASTL DFA type
//
// Requirements:
// 1. The DFA Tag must be of type ForwardCursor::State
//
// Constructors:
// 1. A forward cursor to be adapted, a DFA

template <class ForwardCursor, class DFA>
class lazy_cursor : public forward_cursor<DFA>
{
protected:
  typedef map<typename ForwardCursor::State, safe_state<typename DFA::State> > mapper;
  ForwardCursor       c;
  DFA                 *dfa;
  mapper              mapping;

public:
  typedef forward_cursor<DFA> super;
  typedef typename DFA::State State;
  typedef typename Tag::State Tag;
  
  lazy_cursor(const ForwardCursor &x, DFA &a)
    : super(a), c(x), dfa(&a)
  { }

  void forward() {
    super::forward();
  }

  bool forward(unsigned int a) {
    if (find(a)) {
      forward();
      return true;
    }
    return false;
  }
	
  bool first_transition() {
    c = src_tag();
    if (c.first_transition()) {
      find(c.letter());
      return true;
    }
    return false;
  }

  bool next_transition() {
    c = src_tag();
    c.find(letter());
    if (c.next_transition()) {
      find(c.letter());
      return true;
    }
    return false;
  }
	
  bool find(unsigned int a) {
    if (super::find(a))
      return true;
    c = src_tag();
    if (c.find(a)) {
      typename DFA::State &p = mapping[c.aim()];
      if (p == null_state) {
	p = dfa->new_state();
	dfa->final(p) = c.aim_final();
	dfa->tag(p) = c.aim();
      }
      dfa->set_trans(src(), a, p);
      super::find(a);
      return true;
    }
    return false;
  }
};
     
#endif // ASTL_CLASS_LAZY_CURSOR    
    

