#include <astl.h>
#include <dfa_map.h>
#include <cursors/cursor.h>
#include <cursors/language.h>
#include <cursors/combinatory.h>
#include <cursors/filter.h>
#include <cursors/set_operation.h>
#include <astl_tree.h>
#include <iterator>
#include <fstream>
#include <iostream>
#include <string>
#include <vector>
#include <functional>
#include <cursors/helpers.h>
#include <cursors/ccopy.h>
#include <minimize.h>

typedef DFA_map<Type_alphabet<char>, minimization_tag> DFA;
typedef forward_cursor<DFA> ForwardCursor;

// compile with:
// g++ -Wall -O3 -I../.. anagrammes.cpp

// Usage: a.out dico
// dico contains one word per line
// then for each word entered, the program
// displays all permutations of the word which are in the dictionary

struct anagrames : public unary_function<unsigned int, unsigned int>
{
  string w;
  unsigned int operator() (unsigned int i) const {
    return w[i - 1];
  }
  anagrames(const string &s)
    : w(s) {
    sort(w.begin(), w.end());
  }
};

int main(int argc, char **argv)
{
  DFA dfa1, dfa2;  
  if (argc > 1) {
    ifstream f1;
    f1.open(argv[1], ios::in);
    if (!f1) exit(2);
    tree_build(dfa1, istream_iterator<string>(f1), istream_iterator<string>());
    f1.close();
  }
  cout << "dfa1::Q = " << dfa1.state_count() << " ::T = " << dfa1.trans_count()
       << endl;
  cout << "minimizing...";
  cout.flush();
  acyclic_minimization(dfa1);
  cout << endl << "dfa1::Q = " << dfa1.state_count() << " ::T = " << dfa1.trans_count()
       << endl;

  while(1) {
    string w;
    cout << "Ok" << endl;
    cin >> w;
    cout << "language:" << endl;
    language(ostream_iterator<char>(cerr), 
	     dfirstc(intersectionc(filter_outc(permutationc(w.size()),
					      anagrames(w.c_str())) ,
				   forwardc(dfa1))));
    /* ccopy test: 
    cout << "copy:" << endl;
    DFA::State i = ccopy(dfa2, dfirstc(intersectionc(filter_outc(permutationc(w.size()),
								 anagrames(w.c_str())) ,
						     forwardc(dfa1))));
    if (i != dfa2.null_state) {
      dfa2.initial(i);
      language(ostream_iterator<char>(cerr), dfirstc(forwardc(dfa2)));
    }
    */
  }
  
  cout << "exit." << endl;
  return 0;
}
  
  








