#ifndef ASTL_SET_OPERATION_CURSORS
#define ASTL_SET_OPERATION_CURSORS

// Defines:
// template union_cursor (forward cursor adapter)
// template intersection_cursor (forward cursor adapter)
// template diff_cursor (forward cursor adapter)
// template sym_diff_cursor (forward cursor adapter) (not tested)
// template concatenation_cursor (forward cursor adapter) (not tested)
// template not_cursor (forward cursor adapter) (not tested)

// These adapters are binary adapters: they adapt two forward cursors

// Instanciation parameters:
// 1. ForwardCursor1: left cursor type
// 2. ForwardCursor2: right cursor type
//
// Requirements: 
// 1. Outgoing transitions must be sorted in increasing order
//
// Constructors:
// 1. Two forward cursors

// Implmentations de find  vrifier
// TODO: optimisations
// On risque d'avoir des problmes avec src_final().

#include <list>
#include <vector>
#include <utility>  // pair<>

template <class ForwardCursor> class set_operators;

// Mthode sink a faire
template <class ForwardCursor1, class ForwardCursor2>
class union_cursor
{
protected:
  ForwardCursor1 x1;
  ForwardCursor2 x2;
  bool ok1, ok2;   // ok1,2 == true <=> x1,2 is on a defined state (not the sink state)

public:
  typedef pair<typename ForwardCursor1::State, typename ForwardCursor2::State> State;

  union_cursor(const ForwardCursor1 &x1, const ForwardCursor2 &x2)
    : x1(x1), x2(x2), ok1(true), ok2(true)
  { }

  bool operator== (const union_cursor &x) const {
    return x1 == x.x1 && x2 == x.x2 && ok1 == x.ok1 && ok2 == x.ok2;
  }

  void forward() {
    if (!ok1)  // case (X, q2)
      x2.forward();
    else
      if (!ok2) // case (q1, X)
	x1.forward();
      else      // case (q1, q2), move along transition with smallest letter
	        // possibly both
	if (x1.letter() == x2.letter()) {
	  x1.forward();
	  x2.forward();
	}
	else
	  if (x1.letter() < x2.letter()) {
	    x1.forward();
	    ok2 = false;
	  }
	  else {
	    x2.forward();
	    ok1 = false;
	  }
  }

  bool first_transition() {
    ok1 = ok1 && x1.first_transition();
    ok2 = ok2 && x2.first_transition();
    return (ok1 || ok2);
  }

  bool next_transition() {
    if (ok1)
      if (ok2) {   // case (q1, q2)
	// on prend la transition avec la plus "petite" lettre et on passe  la
	// suivante (si x1.letter() == x2.letter() on passe  la suivante sur les 2)
	unsigned int tmp = x1.letter();
	if (tmp <= x2.letter())
	  ok1 = x1.next_transition();
	if (x2.letter() <= tmp)
	  ok2 = x2.next_transition();
      }
      else       // case (q1, X)
	ok1 = x1.next_transition();
    else
        // case (X, q2) (si ok1 == false, on a forcment ok2 == true)
	ok2 = x2.next_transition();
    
    return (ok1 || ok2);
  }

  unsigned int letter() const {
    if (!ok1) return x2.letter();
    else
      if (!ok2) return x1.letter();
    else
      return min(x1.letter(), x2.letter());
  }

  bool find(unsigned int letter) {
    bool f1 = ok1 && x1.find(letter), f2 = ok2 && x2.find(letter);
    if (f1 || f2) {
      ok1 = f1;
      ok2 = f2;
      return true;
    }
    return false;
  }

  bool forward(unsigned int letter) {
    ok1 = ok1 && x1.forward(letter);
    ok2 = ok2 && x2.forward(letter);
    return (ok1 || ok2);
  }

  // Under construction. Do not use:
  State src() const {
    return make_pair(x1.src(), x2.src());
  }

  // Under construction. Do not use:
  State aim() const {
    return make_pair(x1.aim(), x2.aim());
  }

  bool src_final() const {
    return ((ok1 && x1.src_final()) || (ok2 && x2.src_final()));
  }

  bool aim_final() const {
    return ((ok1 && x1.aim_final()) || (ok2 && x2.aim_final()));
  }
};

template <class ForwardCursor1, class ForwardCursor2>
class intersection_cursor
{
protected:
  ForwardCursor1 x1;
  ForwardCursor2 x2;   

public:
  typedef pair<typename ForwardCursor1::State, typename ForwardCursor2::State> State;
 
  intersection_cursor(const ForwardCursor1 &y1, const ForwardCursor2 &y2)
    : x1(y1), x2(y2)
  { }

  bool sink() const {
    return x1.sink();
  }

  State src() const {
    return State(x1.src(), x2.src());
  }

  State aim() const {
    return State(x1.aim(), x2.aim());
  }

  bool operator== (const intersection_cursor &y) const {
    return x1 == y.x1 && x2 == y.x2;
  }

  void forward() {
    x1.forward();
    x2.forward();
  }

  bool first_transition() {
    if (x1.first_transition() && x2.first_transition())
      while(1) 
	if (x1.letter() < x2.letter()) {
	  if (!x1.next_transition()) return false;
	}
	else
	  if (x2.letter() < x1.letter()) {
	    if (!x2.next_transition()) return false;
	  }
	  else    //  *x1 == *x2 
	    return true;
    return false;
  }

  bool next_transition() {
    if (x1.next_transition() && x2.next_transition())
      while(1) 
	if (x1.letter() < x2.letter()) {
	  if (!x1.next_transition()) return false;
	}
	else
	  if (x2.letter() < x1.letter()) {
	    if (!x2.next_transition()) return false;
	  }
	  else    //  *x1 == *x2 
	    return true;
    return false;
  }

  unsigned int letter() const {
    return x1.letter();
  }

  bool find(unsigned int a) {
    return (x1.find(a) && x2.find(a));
  }

  bool forward(unsigned int a) {
    return x1.forward(a) && x2.forward(a);
  }

  bool src_final() const {
    return (x1.src_final() && x2.src_final());
  }

  bool aim_final() const {
    return (x1.aim_final() && x2.aim_final());
  }
};
  
template <class ForwardCursor1, class ForwardCursor2>
class diff_cursor
{
protected:
  ForwardCursor1         x1;
  mutable ForwardCursor2 x2;  // mutable because methods *_final() are const 
  bool ok2;       // ok2 == true <=> x2 is currently in use

public:
  typedef typename ForwardCursor1::State State;
  
  diff_cursor(const ForwardCursor1 &x1, const ForwardCursor2 &x2)
    : x1(x1), x2(x2), ok2(true)
    { }
  
  bool operator== (const diff_cursor &y) const {
    if (ok2)  
      return y.ok2 && x2 == y.x2 && x1 == y.x1;
    else
      return !y.ok2 && x1 == y.x1;
  }

  unsigned int letter() const {
    return x1.letter();
  }

  void forward() {
    ok2 = ok2 && x2.forward(*x1);
    x1.forward();
  }

  bool first_transition() {
    return x1.first_transition();
  }

  bool next_transition() {
    return x1.next_transition();
  }

  bool find(unsigned int letter) {
    return (x1.find(letter));
  }

  bool forward(unsigned int letter) {
    ok2 = ok2 && x2.forward(letter);
    return x1.forward(letter);
  }

  State src() const {
    return x1.src();
  }

  State aim() const {
    return x1.aim();
  }

  bool src_final() const {
    if (ok2)
      return x1.src_final() && !x2.src_final();
    else
      return x1.src_final();
  }

  bool aim_final() const {
    if (ok2)
      return x1.aim_final() && x2.find(*x1) && !x2.aim_final();
    else
      return x1.aim_final();
  }
};

// Diffrence symmtrique(A1, A2)
// (A1 || A2) \ (A1 && A2)
template <class C1, class C2>   // C1,C2 forward cursor types
class sym_diff_cursor 
  : public diff_cursor<union_cursor<C1, C2>, intersection_cursor<C1, C2> >
{
public:
  typedef diff_cursor<union_cursor<C1, C2>, intersection_cursor<C1, C2> > super; 
  typedef pair<typename C1::State, typename C2::State> State;

  sym_diff_cursor(const C1 &x1, const C2 &x2)
    : super(union_cursor<C1, C2>(x1, x2), intersection_cursor<C1, C2>(x1, x2))
  { }
};


/*
// Version non teste donc bugge :-)
// Pour une version peut-tre plus optimale que la prcdente ( voir):  
template <class ForwardCursor1, class ForwardCursor2>
class sym_diff_cursor
{
protected:
  ForwardCursor1 x1;
  ForwardCursor2 x2;
  bool ok2;       // ok2 == true <=> x2 has no singular value (is a defined
                       // state and is usable)

public:
  sym_diff_cursor(const ForwardCursor1 &x1, const ForwardCursor2 &x2)
    : x1(x1), x2(x2), ok2(true)
  { }

  void forward() {
    ok2 = ok2 && x2.forward(*x1);
    x1.forward();
  }

  bool first_transition() {
    return x1.first_transition();
  }

  bool next_transition() {
    return x1.next_transition();
  }

  // Lettre courante:
  unsigned int operator* () const {
    return *x1;
  }

  bool find(unsigned int letter) {
    return (x1.find(letter));
  }

  bool forward(unsigned int letter) {
    ok2 = ok2 && x2.forward(letter);
    return x1.forward(letter);
  }

  bool src_final() const {
    if (ok2)
      return x1.src_final() && !x2.src_final();
    else
      return x1.src_final();
  }

  bool aim_final() const {
    if (ok2)
      return x1.aim_final() && !x2.aim_final();
    else
      return x1.aim_final();
  }

  bool final() const {
    if (ok2)
      return x1.final() && !x2.final();
    else
      return x1.final();    
  }
};

*/

// Concatnation du langage reconnu par ForwardCursor1 et ForwardCursor2
// Lors de la construction, ForwardCursor2 doit tre positionn sur l'tat
// initial du 2me automate

// Prdicat pour la recherche de la transition d'tiquette minimum grce 
// l'algorithme standard min_element
template <class ForwardCursor>
struct compare_cursor : public binary_function<ForwardCursor, ForwardCursor, bool>
{
  bool operator() (const ForwardCursor &x, const ForwardCursor &y) const {
    return *x < *y;
  }
};

template <class ForwardCursor1, class ForwardCursor2>
class concatenation_cursor
{
protected:
  ForwardCursor1 x1;
  ForwardCursor2 i;
  list<ForwardCursor2> x2;
  bool ok1;
  unsigned int min_letter;
  typedef list<ForwardCursor2> container;

public:
  typedef pair<typename ForwardCursor1::State, vector<typename ForwardCursor2::State> > State;

  concatenation_cursor(const ForwardCursor1 &c1, const ForwardCursor2 &c2)
    : x1(c1), i(c2), ok1(true) { 
    if (x1.src_final()) x2.push_back(i);
  }

  bool sink() const {
    return min_letter == (unsigned int) -1;
  }

  State src() const {
    State result(x1.src(), vector<typename ForwardCursor2::State>());
    result.second.reserve(8);
    for(container::const_iterator i = x2.begin(); i != x2.end(); ++i)
      result.second.push_back((*i).src());
    return result;
  }

  State aim() const {
    State result(x1.aim(), vector<typename ForwardCursor2::State>());
    result.second.reserve(8);
    for(container::const_iterator i = x2.begin(); i != x2.end(); ++i)
      if (!(*i).sink()) result.second.push_back((*i).aim());
    return result;
  }
    

  void forward() {
    for(container::iterator j = x2.begin(); j != x2.end(); )
      if ((*j).letter() == min_letter) (*j++).forward();
      else x2.erase(j++);
    
    if (ok1)
      if (x1.sink()) ok1 = false;
      else
	if (x1.letter() == min_letter) {
	  x1.forward();
	  if (x1.src_final()) x2.push_back(i);
	}
  }

  bool first_transition() {
    min_letter = (unsigned int) -1;  
    if (ok1 && x1.first_transition())
	min_letter = x1.letter();

    for(container::iterator j = x2.begin(); j != x2.end(); ++j)
      if ((*j).first_transition()) {
	if ((*j).letter() < min_letter)
	  min_letter = (*j).letter();
      }

    return min_letter != (unsigned int) -1;  
  }

  bool next_transition() {
    unsigned int next_min = (unsigned int) -1;
    if (ok1 && !x1.sink() && x1.letter() == min_letter) 
      if (x1.next_transition())
	next_min = x1.letter();
  
    for(container::iterator j = x2.begin(); j != x2.end(); ++j)
      if (!(*j).sink() && (*j).letter() == min_letter)
	if ((*j).next_transition()) {
	  if ((*j).letter() < next_min)
	    next_min = (*j).letter();
	}

    min_letter = next_min;
    return min_letter != (unsigned int) -1;
  }
    
  unsigned int letter() const {
    return min_letter;
  }

 
  bool find(unsigned int letter) {
    bool result = ok1 && x1.find(letter);
    for(container::iterator i = x2.begin(); i != x2.end(); ++i)
      result |= (*i).find(letter);
    return result;
  } 
  
  bool forward(unsigned int letter) {
    if (find(letter)) {
      forward();
      return true;
    }
    return false;
  }
  
  bool src_final() const {
    for(container::const_iterator j = x2.begin(); j != x2.end(); ++j)
      if ((*j).src_final()) return true;
    return false;
  }

  bool aim_final() const {
    for(container::const_iterator j = x2.begin(); j != x2.end(); ++j)
      if (!(*j).sink() && (*j).letter() == min_letter && (*j).aim_final()) 
	return true;

    return false;
  }

  bool operator== (const concatenation_cursor &x) const {
    return x1 == x.x1 && x2 == x.x2;
  }
};    


template <class ForwardCursor, unsigned int alphabet_size = 256>
class not_cursor 
{
protected:
  static typename ForwardCursor::Tag sink_tag;
  static const int SRC, AIM;
  ForwardCursor c;
  unsigned int l;
  bool _sink[2];

public:
  typedef not_cursor                    self;
  typedef typename ForwardCursor::State State;
  typedef typename ForwardCursor::Tag   Tag;
  
  not_cursor(const ForwardCursor &x)
    : c(x), l(0) { 
    _sink[SRC] = _sink[AIM] = false;
  }

  bool sink() const {
    return _sink[AIM];
  }

  State src() const {
    return _sink[SRC] ? State() : c.src();
  }

  State aim() const {
    return _sink[AIM] ? State() : c.aim();
  }    

  const Tag& src_tag() const {
    return _sink[SRC] ? sink_tag : c.src_tag();
  }
  
  const Tag& aim_tag() const {
    return _sink[AIM] ? sink_tag : c.aim_tag();
  }    
  
  bool src_final() const {
    return _sink[SRC] || !c.src_final();
  }
  
  bool aim_final() const {
    return _sink[AIM] || !c.aim_final();
  }

  void forward() {
    if (!_sink[SRC])
      if (!_sink[AIM]) 
        c.forward();
      else _sink[SRC] = true;
  }
  
  bool forward(unsigned int a) {
    _sink[SRC] = _sink[SRC] || !c.forward(a);
    return true;
  }
  
  unsigned int letter() const {
    return l;
  }
  
  bool first_transition() {
    l = 0;
    _sink[AIM] = !_sink[SRC] && !c.find(0); 
    return true;
  }
  
  bool next_transition() {
    ++l;
    if (l < max_letter && !_sink[SRC])
      _sink[AIM] = !c.find(l);
    return l < max_letter;
  }

  bool find(unsigned int a) {
    if (!_sink[SRC])
      _sink[AIM] = !c.find(a);
    l = a;
    return true;
  }

  bool operator== (const self &x) const {
    if (_sink[SRC])
      return x._sink[SRC];
    if (x._sink[SRC])
      return false;
    return src() == x.src() && letter() == x.letter();
  }
};    

template <class T, unsigned int ma>
const int not_cursor<T, ma>::SRC = 0;

template <class T, unsigned int ma>
const int not_cursor<T, ma>::AIM = 1;

#endif




